from amino import List, __, _
from amino.lazy import lazy

from ribosome.record import Record, field, maybe_field

from myo.ui.tmux.adapter import Adapter
from myo.ui.tmux.pane import PaneAdapter
from myo.ui.tmux.layout import Layout
from myo.ui.tmux.util import parse_window_id


class Window(Record):
    id = maybe_field(int)
    root = field(Layout)

    @property
    def desc(self):
        id = self.id / ' @{}'.format | ''
        return 'W{}'.format(id)

    @property
    def name(self):
        return self.root.name


class VimWindow(Window):

    @property
    def desc(self):
        return 'V{}'.format(super().desc)


class WindowAdapter(Adapter):

    @lazy
    def panes(self):
        return List.wrap(self.native.panes) / PaneAdapter

    def pane_by_id(self, id: int):
        return self.panes.find(__.id_i.contains(id))

    @lazy
    def id(self):
        return self.native._window_id

    @lazy
    def id_i(self):
        return parse_window_id(self.id)

    @lazy
    def size(self):
        return int(self.native.width), int(self.native.height)

    @property
    def active_pane(self):
        return self.panes.find(_.active)

__all__ = ('WindowAdapter', 'Window')



import re

from amino import Right, Left, Map, Maybe, __, F, Either, List


def parse_int(i):
    return Right(i) if isinstance(i, int) else (
        Right(int(i)) if isinstance(i, str) and i.isdigit() else
        Left('could not parse int {}'.format(i))
    )


def parse_id(value, rex, desc):
    return (
        Maybe(rex.match(str(value)))
        .map(__.group(1))
        .map(int)
        .to_either("could not match {} id {}".format(desc, value))
        .or_else(lambda: parse_int(value)))


def optional_params(m: Map, *keys: str):
    return Map(List.wrap(keys) / (lambda a: (a, m.get(a))))


def bool_params(m, *keys):
    return Map(List.wrap(keys) // (lambda a: m.get(a) / (lambda b: (a, b))))


def view_params(m: Map, *extra):
    return optional_params(m, 'min_size', 'max_size', 'fixed_size', 'position',
                           'weight', 'minimized_size', *extra)


def pane_params(m: Map):
    return view_params(m)


def pane_bool_params(m: Map):
    return bool_params(m, 'pin', 'focus', 'kill')

_py_callback_re = re.compile('^py:(.+)')
_vim_callback_re = re.compile('^vim:(.+)')


def _cb_err(data):
    return 'invalid callback string: {}'.format(data)


def parse_python_callback(data: str):
    return (
        Maybe(_py_callback_re.match(data)) /
        __.group(1) /
        Either.import_path
    ).to_either(_cb_err(data))


def parse_vim_callback(data: str):
    return Left(_cb_err(data))


def parse_callback_spec(data: str):
    return parse_python_callback(data).or_else(F(parse_vim_callback, data))


def amend_options(opt: Map, key: str, value: Maybe):
    return value / (lambda a: opt + (key, a)) | opt

__all__ = ('parse_int', 'optional_params', 'view_params', 'parse_id',
           'parse_callback_spec', 'pane_params')



from amino.test.path import fixture_path, base_dir
from amino import List

from myo.plugins.command.util import assemble_vim_test_line

from integration._support.base import MyoIntegrationSpec
from integration._support.vimtest import vimtest


class VimTestSpec(MyoIntegrationSpec):

    @vimtest
    def vim_test_line(self):
        fname = fixture_path('tmux', 'vim_test', 'test.py')
        self.vim.cmd('noswapfile edit {}'.format(fname))
        self.vim.cursor(5, 0)
        base = base_dir().parent
        target = 'py.test {} -k test_something'.format(fname.relative_to(base))
        assemble_vim_test_line(self.vim).to_maybe.should.contain(List(target))

__all__ = ('VimTestSpec',)



from amino import List, __, _
from amino.test import later

from integration._support.tmux import TmuxIntegrationSpec, DefaultLayoutSpec


class CutSizeSpec(TmuxIntegrationSpec):

    def cut_size(self):
        def check():
            diff = (
                (self._panes.lift(1) & self._panes.lift(2))
                .map2(lambda a, b: abs(a.height - b.height)) |
                10)
            (diff <= 1).should.be.ok
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test', min_size=100)
        self._create_pane('pan2', parent='test', min_size=100)
        self._open_pane('pan1')
        self._open_pane('pan2')
        later(check)


class DistributeSizeSpec(TmuxIntegrationSpec):

    def distribute(self):
        h1 = 8
        check = lambda: (self._sizes / __[1]).last.should.contain(h1)
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test', min_size=5, max_size=h1,
                          weight=1)
        self._create_pane('pan2', parent='test', min_size=10, max_size=40,
                          weight=h1 + 1)
        self._open_pane('pan1')
        self._wait(.1)
        self._open_pane('pan2')
        later(check)

    def default_positions(self):
        sz1, sz2 = 5, 10
        check = lambda: (
            self._panes
            .filter((_.id_i | 0) > 0)
            .sort_by(_.top)
            .map(_.height)
            .drop(1)
            .should.equal(List(sz1, sz2))
        )
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test', weight=1)
        self._wait(.1)
        self._create_pane('pan2', parent='test', fixed_size=sz1)
        self._wait(.1)
        self._create_pane('pan3', parent='test', fixed_size=sz2)
        self._open_pane('pan1')
        self._wait(.1)
        self._open_pane('pan2')
        self._wait(.1)
        self._open_pane('pan3')
        later(check)

    def move(self):
        sz1, sz2 = 5, 10
        check = lambda: (
            self._panes
            .filter((_.id_i | 0) > 0)
            .sort_by(_.top)
            .map(_.height)[:2]
            .should.equal(List(sz1, sz2))
        )
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test', position=1, weight=1)
        self._create_pane('pan2', parent='test', position=0.5, fixed_size=sz1)
        self._create_pane('pan3', parent='test', position=0.7, fixed_size=sz2)
        self._open_pane('pan1')
        self._open_pane('pan2')
        self._open_pane('pan3')
        later(check)


class DefaultLayoutDistributeSizeSpec(DefaultLayoutSpec):

    def _set_vars(self):
        super()._set_vars()
        self.vim_width = 10
        self.vim.vars.set_p('tmux_vim_width', self.vim_width)

    def distribute(self):
        def check():
            widths = self._sizes / __[0]
            target = List(self.vim_width, self.win_width - self.vim_width)
            widths.should.equal(target)
        self._open_pane('make')
        later(check)

    def minimize(self):
        mini = 5
        self._create_pane('pan', parent='root', fixed_size=20,
                          minimized_size=mini, weight=0)
        self._open_pane('pan')
        self.vim.cmd_sync('MyoTmuxMinimize pan')
        self._width(1, mini)

    def minimize_layout(self):
        mini = 5
        self.json_cmd_sync('MyoTmuxCreateLayout lay', parent='root',
                           fixed_size=20, minimized_size=mini, weight=0)
        self._create_pane('pan', parent='lay')
        self._open_pane('lay')
        self.vim.cmd_sync('MyoTmuxMinimize lay')
        self._width(1, mini)


class MinimizeSpec(TmuxIntegrationSpec):

    def _check(self, size):
        later(lambda: self._panes.find(__.id_i.contains(1))
              .map(_.height)
              .should.contain(size))

    def minimize(self):
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', minimized_size=5, fixed_size=10,
                          parent='test')
        self._create_pane('pan2', fixed_size=10, parent='test')
        self._create_pane('pan3', parent='test')
        self._open_pane('pan1')
        self._open_pane('pan2')
        self._open_pane('pan3')
        self._pane_count(4)
        self._check(10)
        self.json_cmd_sync('MyoTmuxMinimize pan1')
        self._check(5)
        self.json_cmd_sync('MyoTmuxRestore pan1')
        self._check(10)

    def minimize_default(self):
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test')
        self._create_pane('pan2', parent='test')
        self._open_pane('pan1')
        self._open_pane('pan2')
        self._pane_count(3)
        self.json_cmd_sync('MyoTmuxMinimize pan1')
        self._check(2)

    def open_or_toggle_pane(self):
        fix = 10
        mini = 5
        cmd = lambda: self.json_cmd_sync('MyoTmuxOpenOrToggle pan')
        self.json_cmd_sync('MyoTmuxCreateLayout lay', parent='root')
        self._create_pane('pan', fixed_size=fix, minimized_size=mini,
                          parent='lay')
        self._create_pane('pan2', parent='lay', weight=1)
        cmd()
        self._wait(.1)
        self._open_pane('pan2')
        self._pane_count(3)
        self._height(1, fix)
        cmd()
        self._height(1, mini)
        cmd()
        self._height(1, fix)

    def open_or_toggle_layout(self):
        fix = 10
        mini = 5
        cmd = lambda: self.json_cmd_sync('MyoTmuxOpenOrToggle lay2')
        self.json_cmd_sync('MyoTmuxCreateLayout lay', parent='root')
        self.json_cmd_sync('MyoTmuxCreateLayout lay2', parent='lay',
                           fixed_size=fix, minimized_size=mini, weight=0)
        self._create_pane('pan', parent='lay2')
        self._create_pane('pan2', parent='lay', weight=1)
        cmd()
        self._wait(.1)
        self._open_pane('pan2')
        self._pane_count(3)
        self._wait(2)
        self._height(1, fix)
        cmd()
        self._height(1, mini)
        cmd()
        self._height(1, fix)

__all__ = ('CutSizeSpec', 'DistributeSizeSpec',
           'DefaultLayoutDistributeSizeSpec')



from libtmux.common import TmuxRelationalObject

from myo.logging import Logging


class Adapter(Logging):

    def __init__(self, native) -> None:
        if not isinstance(native, TmuxRelationalObject):
            raise Exception(
                'passed {!r} to {}'.format(native, self.__class__.__name__))
        self.native = native

    def __repr__(self):
        return 'A({})'.format(self.native)

__all__ = ('Adapter',)



import abc
from typing import Callable

from amino import List, Path, Either, __, L, _, Just, Maybe, Map
from amino.task import Task
from amino.lazy import lazy

from ribosome import NvimFacade
from ribosome.machine.state import RunScratchMachine
from ribosome.machine import Nop, Message

from myo.logging import Logging
from myo.output.data import ParseResult
from myo.output.machine import OutputMachine


class OutputHandler(Logging, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def parse(self, output: List[str], errfile: Path) -> Task[ParseResult]:
        ...

    @abc.abstractmethod
    def display(self, result: ParseResult, jump: Maybe[Callable]
                ) -> Task[List[Message]]:
        ...


class CustomOutputHandler(OutputHandler):

    def __init__(self, vim: NvimFacade, handler: Callable[[str], ParseResult]
                 ) -> None:
        self.vim = vim
        self.handler = handler

    def parse(self, output: List[str], errfile: Path):
        return Task.call(self.handler, output)

    def display(self, result: ParseResult, options: Map[str, str]):
        ctor = L(OutputMachine)(self.vim, _, result, _, options)
        size = self.vim.vars.p('scratch_size') | 10
        return Task.now(Just(RunScratchMachine(
            ctor, options=Map(use_tab=False, size=Just(size), wrap=True)).pub))


class VimCompiler(OutputHandler):

    def __init__(self, vim: NvimFacade) -> None:
        self.vim = vim

    def parse(self, output: List[str], errfile: Path):
        r = ParseResult(head='errorformat')
        return (Task.call(self.vim.cmd_sync, 'cgetfile {}'.format(errfile))
                .replace(r))

    def display(self, result, jump):
        copen = Task.call(self.vim.cmd_sync, 'copen')
        cfirst = Task.call(self.vim.cmd_sync, 'cfirst')
        return (copen + cfirst).replace(Just(Nop()))


class Parsing(CustomOutputHandler):

    def __init__(self, vim: NvimFacade, langs: List[str]) -> None:
        self.vim = vim
        self.langs = langs

    @lazy
    def parsers(self):
        mod = 'myo.output.parser'
        def create(lang):
            return Either.import_name('{}.{}'.format(mod, lang),
                                      'Parser').to_list
        return self.langs // create / (lambda a: a())

    def parse(self, output: List[str], errfile: Path):
        events = self.parsers // __.events(output)
        return Task.now(ParseResult(head='parsed', events=events,
                                    langs=self.langs))

__all__ = ('CustomOutputHandler', 'VimCompiler', 'Parsing')



from integration._support.tmux import DefaultLayoutSpec, TmuxIntegrationSpec

from amino.test import later
from amino import _


class OpenSpec(TmuxIntegrationSpec):

    def open(self):
        self.json_cmd('MyoTmuxCreateLayout lay', parent='root')
        self._create_pane('pan', parent='lay')
        self._open_pane('pan')
        self._pane_count(2)


class ClosePaneSpec(DefaultLayoutSpec):

    def _close(self):
        self._open_pane('make')
        self._pane_count(2)
        self.vim.cmd('MyoTmuxClosePane make')
        self._pane_count(1)
        self.vim.cmd('MyoTmuxClosePane make')
        self._wait(0.1)
        self._pane_count(1)
        self._open_pane('make')
        self._pane_count(2)
        self.vim.cmd('MyoTmuxClosePane make')
        self._pane_count(1)

    def _auto_close(self):
        self._open_pane('make')
        self._pane_count(2)
        self.vim.doautocmd('VimLeave')
        self._pane_count(1)


class ShowSpec(TmuxIntegrationSpec):

    def output(self):
        def check():
            self._log_out.should_not.be.empty
        self.vim.cmd('MyoTmuxShow')
        later(check)


class PinSpec(TmuxIntegrationSpec):

    def open(self):
        check = lambda: self._panes.length.should.equal(3)
        def check2():
            (self._panes.find(_.height == 2) / _.active).should.contain(True)
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test', pin=True)
        self._create_pane('pan2', parent='test', focus=True, fixed_size=2)
        self._open_pane('pan2')
        later(check)
        later(check2)


class FocusSpec(TmuxIntegrationSpec):

    def open_no_focus(self):
        check = lambda i: (self._panes.lift(i) / _.active).should.contain(True)
        self._create_pane('pan1')
        self._open_pane('pan1')
        self._pane_count(2)
        later(check, 0)
        self.vim.cmd_sync('MyoTmuxFocus pan1')
        later(check, 1)

    def open_focus(self):
        check = lambda: (self._panes.lift(1) / _.active).should.contain(True)
        self._create_pane('pan1', focus=True)
        self._open_pane('pan1')
        later(check)


class OpenLayoutSpec(TmuxIntegrationSpec):

    def _pre_start(self):
        super()._pre_start()
        self.json_cmd('MyoTmuxCreateLayout lay', parent='root')
        self._create_pane('pan1', parent='lay')

    def open(self):
        self._open_pane('lay')
        self._pane_count(2)

    def toggle(self):
        self.json_cmd('MyoTmuxOpenOrToggle lay')
        self._pane_count(2)

    def minimize(self):
        self.json_cmd('MyoTmuxCreateLayout lay2', parent='root')
        self._create_pane('pan2', parent='lay2')
        self._open_pane('lay')
        self._open_pane('lay2')
        self._pane_count(3)
        self.json_cmd('MyoTmuxMinimize lay2')
        later(lambda: (self._pane_with_id(2) / _.width).should.contain(2))

    def open_twice(self):
        self._create_pane('pan', parent='lay')
        self.vim.cmd_sync('MyoTmuxOpen lay')
        self._pane_count(2)
        self.vim.cmd_sync('MyoTmuxOpen lay')
        self._wait(1)
        self._pane_count(2)

__all__ = ('ClosePaneSpec', 'ShowSpec')



from ribosome.machine import message, json_message

TmuxRunCommand = json_message('TmuxRunCommand', 'command')
TmuxRunShell = json_message('TmuxRunShell', 'shell')
TmuxRunCommandInShell = message('TmuxRunCommandInShell', 'shell', 'command',
                                'options')
TmuxRunLineInShell = json_message('TmuxRunLineInShell', 'shell')
TmuxCreateSession = json_message('TmuxCreateSession')
TmuxSpawnSession = message('TmuxSpawnSession', 'id')
TmuxCreateLayout = json_message('TmuxCreateLayout', 'name')
TmuxCreatePane = json_message('TmuxCreatePane', 'name')
TmuxOpen = json_message('TmuxOpen', 'name')
TmuxClosePane = message('TmuxClosePane', 'name')
TmuxFindVim = message('TmuxFindVim')
TmuxLoadDefaults = message('TmuxLoadDefaults')
TmuxInfo = message('TmuxInfo')
TmuxParse = json_message('TmuxParse')
TmuxPack = message('TmuxPack')
TmuxPostOpen = json_message('TmuxPostOpen', 'ident')
TmuxFixFocus = message('TmuxFixFocus', 'pane')
TmuxFocus = message('TmuxFocus', 'pane')
TmuxMinimize = json_message('TmuxMinimize', 'pane')
TmuxRestore = json_message('TmuxRestore', 'pane')
TmuxToggle = json_message('TmuxToggle', 'pane')
TmuxOpenOrToggle = json_message('TmuxOpenOrToggle', 'pane')
SetCommandLog = message('SetCommandLog', 'cmd_ident', 'pane_ident')

StartWatcher = message('StartWatcher')
QuitWatcher = message('QuitWatcher')
WatchCommand = message('WatchCommand', 'command', 'pane')

__all__ = ('TmuxRunCommand', 'TmuxOpen', 'TmuxCreatePane',
           'TmuxCreateSession', 'TmuxCreateLayout', 'TmuxCreateLayout',
           'TmuxSpawnSession', 'TmuxFindVim', 'TmuxLoadDefaults',
           'TmuxClosePane', 'TmuxRunCommandInShell', 'TmuxRunLineInShell',
           'TmuxRunShell', 'StartWatcher', 'WatchCommand', 'QuitWatcher',
           'TmuxParse', 'SetCommandLog', 'TmuxPack', 'TmuxPostOpen',
           'TmuxFixFocus', 'TmuxMinimize', 'TmuxRestore', 'TmuxToggle',
           'TmuxFocus', 'TmuxOpenOrToggle')



from ribosome import NvimFacade

from amino import _, L, List, Right, Left


def vimtest_position(vim: NvimFacade):
    fn_mod = vim.vars('g:test#filename_modifier') | ':.'
    line, col = vim.window.cursor
    return vim.call('expand', '%{}'.format(fn_mod)) / (
        lambda f: dict(file=f, line=line, col=col)
    )


def assemble_vim_test_line(vim: NvimFacade):
    def build(exe, bpos):
        fmt = lambda a: '{} {}'.format(exe, a)
        return bpos / fmt
    def args(runner, pos):
        exe = vim.call('test#{}#executable'.format(runner))
        bpos = (vim.call('test#base#build_position', runner, 'nearest', pos) /
                List.wrap)
        return (exe & bpos).map2(build)
    def runner(pos):
        chk_res = lambda a: (
            Right(a) if isinstance(a, str) else Left('no runner found'))
        return (
            vim.call('test#determine_runner', pos['file']) //
            chk_res //
            L(args)(_, pos)
        )
    return vimtest_position(vim) // runner

__all__ = ('assemble_vim_test_line',)



from ribosome.machine import may_handle, Error, RunTask, handle
from ribosome.machine.base import io
from ribosome.machine.transition import Fatal

import amino
from amino import __, F, L, _, Right, Just, List, Left

from myo.state import MyoComponent, MyoTransitions
from myo.plugins.core.dispatch import VimDispatcher
from myo.plugins.core.message import StageI, Initialized, ParseOutput
from myo.output import VimCompiler, CustomOutputHandler, Parsing
from myo.util import parse_callback_spec
from myo.output.machine import ResultAdapter


class CoreTransitions(MyoTransitions):

    @may_handle(StageI)
    def stage_1(self):
        return Initialized().pub.at(1)

    @may_handle(Initialized)
    def initialized(self):
        started = io(__.vars.set_p('started', True))
        return (self.data.set(initialized=True) + VimDispatcher(self.vim),
                started)

    @may_handle(Error)
    def error(self):
        m = self.msg.message
        handler = (F(self.log.caught_exception, 'transition') if
                   amino.development and isinstance(m, Exception) else
                   self.log.error)
        handler(m)

    @handle(ParseOutput)
    def parse_output(self):
        opt = self.msg.options
        filters = opt.get('filters') | List()
        def handle(parser):
            def dispatch(result):
                adapter = ResultAdapter(self.vim, result, filters)
                return parser.display(adapter, opt)
            return RunTask(parser.parse(self.msg.output, self.msg.path) //
                           dispatch)
        return self._error_handler(self.msg.command).join / handle

    def _error_handler(self, cmd):
        langs = self.msg.options.get('langs') / List.wrap | cmd.langs
        parser = self.msg.options.get('parser').or_else(cmd.parser)
        return (
            (parser // self._special_error_handler)
            .or_else(
                parser //
                parse_callback_spec /
                __.map(L(CustomOutputHandler)(self.vim, _))
            ).or_else(self._langs_parsing(langs))
        ).lmap(Fatal)

    def _special_error_handler(self, spec: str):
        return (Right(Right(VimCompiler(self.vim)))
                if spec == 'compiler' else
                Left('no special error handler found'))

    def _langs_parsing(self, langs):
        return langs.empty.no.either('command has no langs',
                                     Just(Parsing(self.vim, langs)))


class Plugin(MyoComponent):
    Transitions = CoreTransitions

    def new_state(self):
        pass

__all__ = ('Plugin',)
